/*
* File WilsonBalding.java
*
* Copyright (C) 2010 Remco Bouckaert remco@cs.auckland.ac.nz
*
* This file is part of BEAST2.
* See the NOTICE file distributed with this work for additional
* information regarding copyright ownership and licensing.
*
* BEAST is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2
* of the License, or (at your option) any later version.
*
*  BEAST is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with BEAST; if not, write to the
* Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
* Boston, MA  02110-1301  USA
*/
/*
 * WilsonBalding.java
 *
 * Copyright (C) 2002-2006 Alexei Drummond and Andrew Rambaut
 *
 * This file is part of BEAST.
 * See the NOTICE file distributed with this work for additional
 * information regarding copyright ownership and licensing.
 *
 * BEAST is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 *  BEAST is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with BEAST; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
 * Boston, MA  02110-1301  USA
 */

/* GRJ, Apr/May 2020: I edited code from the BADTRIP package
 */

package pomocov;

import java.util.List;

import beast.core.Description;
import beast.util.Randomizer;

/**
 * WILSON, I. J. and D. J. BALDING, 1998  Genealogical inference from microsatellite data.
 * Genetics 150:499-51
 * http://www.genetics.org/cgi/ijlink?linkType=ABST&journalCode=genetics&resid=150/1/499
 */
@Description("Implements a move similar to the one proposed by WILSON and BALDING 1998, just picking a random non-root node and assigning another compatible random parent.")
public class WilsonBaldingTransmission extends TreeOperatorTransmission {

    @Override
    public void initAndValidate() {
    }

    /**
     * override this for proposals,
     *
     * @return log of Hastings Ratio, or Double.NEGATIVE_INFINITY if proposal should not be accepted *
     */
    @Override
    public double proposal() {
        double logHR = treeInput.get(this).getUnobservedHosts() ? wilsonBaldingTransmissionUC() : wilsonBaldingTransmission();
        treeInput.get(this).checkTree();
        return logHR;
    }


    boolean newParentOK(NodePoMo node, NodePoMo newParent) {
        boolean ok;
        ok = !newParent.equals(node);
        ok = ok  &&  !newParent.equals(node.getParent());
        double origH = node.getOriginHeight();
        ok = ok  &&  newParent.getEnd() < origH;
        ok = ok  &&  origH < newParent.getHeight();
        return ok;
    }


    // GRJ. Like wilsonBaldingTransmission(), but originHeight, not height, is checked against new parent
    private double wilsonBaldingTransmissionUC() {
        TreePoMo tree = treeInput.get(this);
        assert tree.getExtraList().size() == 0;
        // choose a random node avoiding root
        final int nodeCount = tree.getNodeCount();
        NodePoMo [] nodes = tree.getNodesAsArray();
        NodePoMo node = null;
        do {
            int nodeNr = Randomizer.nextInt(nodeCount);
            node = nodes[nodeNr];
        } while (node.isRoot());
        assert node != null;

        // count candidate new parents, choose one if any to choose from
        int nNewParents = 0;
        for (NodePoMo nP : nodes) {
            if (newParentOK(node, nP)) {
                nNewParents++;
            }
        }
        if (nNewParents == 0) {
            return Double.NEGATIVE_INFINITY;
        }
        NodePoMo newParent = null;
        do {
            int nodeNr = Randomizer.nextInt(nodeCount);
            newParent = nodes[nodeNr];
        } while (!newParentOK(node, newParent));
        assert newParent != null;
        // go for it
        node.getParent().removeChild(node);
        node.setParent(newParent);
        newParent.addChild(node);
        newParent.sortChildren();
        return 0.0;
    }





    private double wilsonBaldingTransmission() {
    	//System.out.println("Start Operator WilsonBaldingTransmission ");
        TreePoMo tree = treeInput.get(this);

        // choose a random node avoiding root
        final int nodeCount = tree.getNodeCount();
        NodePoMo i;
        int nodeNr;
        List<NodePoMo> extras=tree.getExtraList();
        do {
        	//sample random node
        	nodeNr = Randomizer.nextInt(nodeCount);
            if (nodeNr>=extras.size()){
            	i=tree.getNodesAsArray()[nodeNr-extras.size()];
            }else{
            	i=extras.get(nodeNr);
            }
        } while (i.isRoot());
        final NodePoMo p = i.getParent();

        // choose another random node to insert i above
        NodePoMo j;
        //NodePoMo jP;

        // make sure that the target j is a possible parent
        do {
        	//sample random node
        	nodeNr = Randomizer.nextInt(nodeCount);
            if (nodeNr>=extras.size()){
            	j=tree.getNodesAsArray()[nodeNr-extras.size()];
            }else{
            	j=extras.get(nodeNr);
            }
            //j = tree.getNode(Randomizer.nextInt(nodeCount));
            //jP = j.getParent();
        } while (j.equals(i)||j.getHeight()<i.getHeight()||j.getEnd()>i.getHeight());
        
        p.removeChild(i);
        i.setParent(j);
        j.addChild(i);
        j.sortChildren();
        //System.out.println("Operator WilsonBaldingTransmission "+i.getID()+" "+j.getID()+" "+p.getID());
        return 0.0;
        
    }


} // class WilsonBalding
