
/* GRJ, Apr/May 2020
 */


package pomocov.utils;

public class TransmissionDistributions {

    // The scale values are in days

    static final double withinHostShape1 = 6.0;
    static final double withinHostShape2 = 3.0;
    static final double withinHostScale = 10.0;

    static final double [] betweenHostShape1 = {10.0, 5.5, 4.0};
    static final double [] betweenHostShape2 = {0.75, 3.1, 4.7};
    static final double [] betweenHostScale  = {2.0, 9.7, 14.8};

    static final double uLargeShapeProduct = 15.0;
    static final double uLargeWithinScale = 5.0;
    static final double uLargeBetweenScale = 0.13;


    // log-density for transmission in tree prior
    public static  double withinHostLogPrior(double x) {
        return logBurrDensity(x, withinHostShape1, withinHostShape2, withinHostScale);
    }

    // log-density for time for u (=0,1,2,..) unobserved hosts
    public static double betweenHostsLogPrior(int u, double g, double h) {
        double sh1;
        double sh2;
        double scale;
        if (u < 3) {
            sh1 = betweenHostShape1[u];
            sh2 = betweenHostShape2[u];
            scale = betweenHostScale[u];
        } else {
            sh2 = u + 3.0;
            sh1 = uLargeShapeProduct / sh2;
            scale = uLargeWithinScale * (u + Math.pow(0.85, u)) + uLargeBetweenScale * (u+1);
        }
        return logBurrDensity(g - h, sh1, sh2, scale);
    }


    public static double unobservedCountLogPrior(int u, double ucMean) {
        if (ucMean < 1e-8) {
            return 0.0;
        }
        double p = 1.0 / (1.0 + ucMean);
        return u * Math.log(1.0 - p) - Math.log(p);
    }


    // for dividing time between observed hosts given number of unobserved
    public static double medianWithinHostTime() {
        double sh1 = withinHostShape1;
        double sh2 = withinHostShape2;
        double scale = withinHostScale;
        return burrMedian(sh1, sh2, scale);
    }

    // for dividing time between observed hosts given number of unobserved, also for initialising tree
    public static double medianDelayTime() {
        double sh1 = betweenHostShape1[0];
        double sh2 = betweenHostShape2[0];
        double scale = betweenHostScale[0];
        return burrMedian(sh1, sh2, scale);
    }


    // for initialising tree
    public static int sensibleUnobservedCount(double originH, double infH) {
        double sh1 = betweenHostShape1[1];
        double sh2 = betweenHostShape2[1];
        double scale = betweenHostScale[1];
        double median =  burrMedian(sh1, sh2, scale);
        return (int)Math.floor(0.5 +  (originH - infH) / median);
    }


    private static double logBurrDensity(double x, double sh1, double sh2, double scale) {
        double logBD = 0.0;
        // f(x) = (a b (x/s)^b) / (x [1 + (x/s)^b]^(a + 1))
        logBD += Math.log(sh1 * sh2 / x);
        logBD += sh2 * Math.log(x / scale);
        logBD -= (sh1 + 1.0) * Math.log(1.0 + Math.pow(x / scale, sh2));
        return logBD;
    }


    private static double burrMedian(double sh1, double sh2, double scale) {
        return scale * Math.pow((Math.pow(2.0, 1.0 / sh1) - 1.0), 1.0 / sh2);
    }


    private static double burrQuantile(double y, double sh1, double sh2, double scale) {
        return scale * Math.pow((Math.pow((1.0 - y), -1.0 / sh1) - 1.0), 1.0 / sh2);
    }




    private static void insistFloatsClose(double x, double y) {
        assert Math.abs(x - y) < 1e-8;
    }


    public static void sanityChecks() {
        System.err.println("sanityChecks");
        double [] xs = {0.1,0.5,1,2,4,8,16,32};
        double [] logbds = {-4.13636631680633, -2.55557963686776, -1.95024360320815, -1.58469429854318,
                -1.93037582813674, -3.56242155666029, -6.31842927490516, -9.57699591153287};
        for (int i= 0; i < logbds.length; i++) {
            insistFloatsClose(logbds[i], logBurrDensity(xs[i], 2.0, 2.0, 5.0));
        }
        for (int u = 0; u <8; u++) {
            double dens = Math.exp(betweenHostsLogPrior(u, 5*u+.1, 0.0));
            double guess = 1.0 / Math.sqrt(25*u+0.15);
            assert 0.8 < dens/guess;
            assert dens/guess < 1.2;
        }
        assert medianDelayTime() < 0.25;
        assert medianDelayTime() > 0.02;
        assert medianWithinHostTime() < 7.0;
        assert medianWithinHostTime() > 3.0;

    }




}

/*
print(paste0(log(dburr(c(0.1,0.5,1,2,4,8,16,32), 2, 2, scale=5)), ", ", collapse=""), digits=10)
[1] "-4.13636631680633, -2.55557963686776, -1.95024360320815, -1.58469429854318, -1.93037582813674, -3.56242155666029, -6.31842927490516, -9.57699591153287, "
*/